/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  bmp2txt - convert a monochrome BMP to text                       */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>

int main(int argc, char **argv)
{
    size_t sz;
    long lsz;
    unsigned char *buf;
    unsigned char *p;
    FILE *fp;
    FILE *fq;
    unsigned long width;
    unsigned long height;
    unsigned long start;
    unsigned int bpp;
    int x;
    int y;
    int z;

    if (argc <= 2)
    {
        printf("usage: bmp2txt <bmp file> <text file>\n");
        return (EXIT_FAILURE);
    }
    fp = fopen(*(argv + 1), "rb");
    if (fp == NULL)
    {
        printf("failed to open %s for reading\n", *(argv + 1));
        return (EXIT_FAILURE);
    }
    fseek(fp, 0, SEEK_END);
    lsz = ftell(fp);
    sz = (size_t)lsz;
    if (sz != lsz)
    {
        printf("input file too big\n");
        return (EXIT_FAILURE);
    }
    rewind(fp);
    buf = malloc(sz);
    if (buf == NULL)
    {
        printf("insufficient memory\n");
        return (EXIT_FAILURE);
    }
    fread(buf, 1, sz, fp);
    width = ((unsigned long)buf[0x15] << 24)
             | ((unsigned long)buf[0x14] << 16)
             | ((unsigned long)buf[0x13] << 8)
             | buf[0x12];
    height = ((unsigned long)buf[0x19] << 24)
              | ((unsigned long)buf[0x18] << 16)
              | ((unsigned long)buf[0x17] << 8)
              | buf[0x16];
    bpp = ((unsigned int)buf[0x1d] << 8)
           | buf[0x1c];
    if (bpp != 1)
    {
        printf("must be monochrome\n");
        return (EXIT_FAILURE);
    }
    fq = fopen(*(argv + 2), "w");
    if (fq == NULL)
    {
        printf("failed to open %s for writing\n", *(argv + 2));
        return (EXIT_FAILURE);
    }
    start = ((unsigned long)buf[0xd] << 24)
             | ((unsigned long)buf[0xc] << 16)
             | ((unsigned long)buf[0xb] << 8)
             | buf[0xa];
    p = buf + start;
    for (y = height - 1; y >= 0; y--)
    {
        for (x = 0; x < width / 8; x++)
        {
            int c;

            c = p[y * width/8 + x];
            for (z = 7; z >= 0; z--)
            {
                 if (((c >> z) & 1) != 0)
                 {
                     fprintf(fq, " ");
                 }
                 else
                 {
                     fprintf(fq, "X");
                 }
             }
        }
        fprintf(fq, "\n");
    }
    fclose(fq);
    fclose(fp);
    return (0);
}
