/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  objrepro -  generate repro statements from text records in       */
/*  object code.                                                     */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define REPRO "  REPRO             " \
              "                    " \
              "                    " \
              "                    "

#define INMAGIC1 16 /* offset into TXT record that we have data */
#define INMAGIC2 56 /* number of bytes of useful data in a text record */
#define RECLEN 80 /* length of a card */
#define OUTUSE 72 /* how many bytes we can actually use in the output record */

int main(int argc, char **argv)
{
    FILE *fp;
    FILE *fq;
    char outbuf[RECLEN];
    int tot; /* total written to output record so far */
    int remain; /* how much data we have remaining in the output buffer */
    char buf[RECLEN];

    if (argc <= 2)
    {
        printf("usage: objrepro <in> <out>\n");
        return (EXIT_FAILURE);
    }
    fp = fopen(*(argv + 1), "rb");
    if (fp == NULL)
    {
        printf("failed to open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }
    fq = fopen(*(argv + 2), "wb");
    if (fq == NULL)
    {
        printf("failed to open %s\n", *(argv + 2));
        return (EXIT_FAILURE);
    }
    /* we will always get a full text record */
    while (fread(buf, sizeof buf, 1, fp) == 1)
    {
        if (memcmp(buf, "\x02TXT", 4) == 0)
        {
            /* we have an implicit assumption here that we will always
               pad the data out to the full 72-byte length */
            remain = OUTUSE - tot;
            if (remain <= INMAGIC2)
            {
                memcpy(outbuf + tot, buf + INMAGIC1, remain);
                fwrite(REPRO, sizeof REPRO - 1, 1, fq);
                fwrite(outbuf, sizeof outbuf, 1, fq);
                memcpy(outbuf, buf + INMAGIC1 + remain, INMAGIC2 - remain);
                tot = INMAGIC2 - remain;
                fwrite("        ", 8, 1, fq); /* blank sequence number */
            }
            else
            {
                memcpy(outbuf + tot, buf + INMAGIC1, INMAGIC2);
                tot = tot + INMAGIC2;
            }
        }
    }
    return (0);
}

#if 0

ignore the fact that this is silly data due to unrelated issue

000050  02E3E7E3 40000000 40400038 40400001  .TXT ...  ..  ..
000060  4040D9C5 D7D9D640 40404040 40404040    REPRO
000070  40404040 40404040 40404040 40404040
000080  40404040 40404040 40404040 40404040
000090  40404040 40404040 F0F0F0F0 F0F0F0F2          00000002

#endif
