/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  uudecode - decode a uuencoded file                               */
/*                                                                   */
/*  This program should work on both ASCII and EBCDIC platforms.     */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CODECHARS \
  "`!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"

static FILE *inf;
static char outfnm[FILENAME_MAX];
static FILE *outf;
static char buf[100];
static unsigned char value[256];

int main(int argc, char **argv)
{
    static unsigned char *special = (unsigned char *)CODECHARS;
    int x = 0;
    int len;
    unsigned char *p;
    unsigned long y;
    char *z;

    if (argc <= 1)
    {
        /* use defaults */
    }
    else if ((argc == 3) && (strcmp(argv[1], "-o") == 0))
    {
        strncpy(outfnm, argv[2], sizeof outfnm);
    }
    else
    {
        printf("usage: uudecode [-o outfile]\n");
        return (EXIT_FAILURE);
    }

    inf = stdin;

    fgets(buf, sizeof buf, inf);
    if (strncmp(buf, "begin", 5) != 0)
    {
        printf("malformed input - no begin\n");
        return (EXIT_FAILURE);
    }
    z = strchr(buf, '\n');
    if (z != NULL)
    {
        *z = '\0';
    }
    z = strchr(buf, ' ');
    if (z != NULL)
    {
        z = strchr(z + 1, ' ');
        if (z != NULL)
        {
            z++;
        }
    }
    if (z == NULL)
    {
        printf("malformed begin line\n");
        return (EXIT_FAILURE);
    }
    if (outfnm[0] == '\0')
    {
        strncpy(outfnm, z, sizeof outfnm);
    }
    outfnm[sizeof outfnm - 1] = '\0';
    outf = fopen(outfnm, "wb");
    if (outf == NULL)
    {
        printf("unable to open output file %s\n", outfnm);
        return (EXIT_FAILURE);
    }

    while (special[x] != '\0')
    {
        value[special[x]] = (unsigned char)x;
        x++;
    }

    while (fgets(buf, sizeof buf, inf) != NULL)
    {
        if (strncmp(buf, "begin", 5) == 0) continue;
        if (strcmp(buf, "`\n") == 0) break;
        if (strcmp(buf, "end\n") == 0) break;
        p = (unsigned char *)buf;
        len = value[p[0]];
        /* printf("len is %d\n", len); */
        p++;
        for (x = 0; x < len;)
        {
            y = value[*p];
            p++;
            y = (y << 6) | value[*p];
            p++;
            y = (y << 6) | value[*p];
            p++;
            y = (y << 6) | value[*p];
            p++;
            putc((unsigned char)((y >> 16) & 0xff), outf);
            x++;
            if (x >= len) break;
            putc((unsigned char)((y >> 8) & 0xff), outf);
            x++;
            if (x >= len) break;
            putc((unsigned char)(y & 0xff), outf);
            x++;
        }
    }

    return (0);
}
