/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  uuencode - encode a file                                         */
/*                                                                   */
/*  This program should work on both ASCII and EBCDIC platforms.     */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CODECHARS \
  "`!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"

static FILE *inf;
static FILE *outf;
static unsigned char buf[45];
static char *special = CODECHARS;

int main(int argc, char **argv)
{
    unsigned long y;
    int n;
    int i;

    if (argc != 3)
    {
        printf("usage: uuencode <infile> <remotefile>\n");
        return (EXIT_FAILURE);
    }

    inf = fopen(*(argv + 1), "rb");
    if (inf == NULL)
    {
        printf("failed to open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }

    outf = stdout;
    fprintf(outf, "begin 0 %s\n", *(argv + 2));

    while ((n = fread(buf, 1, sizeof buf, inf)) > 0)
    {
        if (n < sizeof buf)
        {
            /* we only need 1 or 2 NUL bytes, but
               it's simpler to clear the whole buffer */
            memset(buf + n, '\0', sizeof buf - n);
        }
        putc(special[n], outf);
        if ((n % 3) != 0)
        {
            n = ((n / 3) * 3) + 3;
        }
        for (i = 0; i < n; i += 3)
        {
            y = (buf[i] << 16)
                | (buf[i + 1] << 8)
                | buf[i + 2];
            putc(special[((y >> 18) & 0x3f)], outf);
            putc(special[((y >> 12) & 0x3f)], outf);
            putc(special[((y >> 6) & 0x3f)], outf);
            putc(special[(y & 0x3f)], outf);
        }
        fprintf(outf, "\n");
    }
    fprintf(outf, "`\nend\n");

    return (0);
}
