/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  vbsanal - analyze a V(B)(S) file                                 */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static unsigned char bdw[4];
static unsigned char *rdw;
static unsigned char buf[70000];
static long offset = 0;
static long maxoff = 0;
static unsigned int maxlen = 0;
static unsigned int len;
static unsigned int rdwlen;
static unsigned int rdwseg;
static unsigned int ret;

int main(int argc, char **argv)
{
    FILE *fp;

    if (argc <= 1)
    {
        printf("usage: vbsanal <vbs file>\n");
        printf("shows BDW + SDW\n");
        printf("SDW of 0 is complete record\n");
        printf("SDW of 1 is first segment\n");
        printf("SDW of 2 is last segment\n");
        printf("SDW of 3 is middle segement\n");
        return (EXIT_FAILURE);
    }

    fp = fopen(*(argv + 1), "rb");
    if (fp == NULL)
    {
        printf("can't open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }

    while (1)
    {
        ret = fread(bdw, 1, 4, fp);
        if (ret == 0) break;
        if (ret < 4)
        {
            printf("only partial bdw at %ld\n", offset);
            break;
        }
        if ((bdw[2] != 0) || (bdw[3] != 0))
        {
            printf("last 2 bytes of BDW non-zero at %ld\n", offset);
            break;
        }
        len = bdw[0] << 8 | bdw[1];
        if (len <= 8)
        {
            printf("bdw too small (%d) at offset %ld\n", len, offset);
            break;
        }
        if (len > maxlen)
        {
            maxlen = len;
            maxoff = offset;
        }
        printf("bdw at offset %ld is %u\n", offset, len);
        len -= 4;
        ret = fread(buf, 1, len, fp);
        if (ret != len)
        {
            printf("only read %u bytes instead of %u at offset %ld\n",
                   ret, len, offset);
            break;
        }
        rdw = buf;
        while (1)
        {
            rdwlen = rdw[0] << 8 | rdw[1];
            rdwseg = rdw[2] << 8 | rdw[3];
            printf(" rdw len is %u, rdw seg is %04X\n", rdwlen, rdwseg);
            if ((rdw - buf + rdwlen) > len)
            {
                printf("rdw is too big for record\n");
                exit(EXIT_FAILURE);
            }
            rdw = rdw + rdwlen;
            if (rdw == (buf + len))
            {
                break;
            }
        }
        offset += (len + 4);
    }
    printf("max BDW was %u at offset %ld\n", maxlen, maxoff);
    return (0);
}
